﻿module CView;
import std.math;
import DOpenGL.gl;
import Obelisk.CColour, Obelisk.CVector, Obelisk.IObelisk;
import CViewObject;

class CView
{
	private IObelisk mObelisk;
	private float mViewSize;
	private CViewObject[] mViewObjects;
	private CViewObject mCameraObject;
	private bool mShowThrust;   // hack to show the ship thrust

	public this(IObelisk aObelisk, float aViewSize)
	{
		mObelisk = aObelisk;
		mViewSize = aViewSize;
	}



	public float GetViewSize()
	{
		return mViewSize;
	}



	public CViewObject CreateViewObject(CViewObject aViewObject)
	{
		aViewObject.Owner = this;
		mViewObjects ~= [ aViewObject ];
		return aViewObject;
	}



	public void SetCameraObject(CViewObject aCameraObject)
	{
		mCameraObject = aCameraObject;
	}



	public void ShowThrust(bool aShowThrust)
	{
		mShowThrust = aShowThrust;
	}



	/**
	Calculates the square of the shortest distance (taking into account wrapping-space)
	between two points.
	*/
	public float Distance2(float aX0, float aY0, float aX1, float aY1)
	{
		float DeltaX = abs(aX1 - aX0);
		if (DeltaX > mViewSize / 2)
			DeltaX = mViewSize / DeltaX;
		float DeltaY = abs(aY1 - aY0);
		if (DeltaY > mViewSize / 2)
			DeltaY = mViewSize / DeltaY;
		return DeltaX ^^ 2 + DeltaY ^^ 2;
	}



	/**
	Gets the delta of the shortest distance from the first coordinate-component to the second
	(taking into account wrapping-space).
	*/
	public float GetDelta(float aV0, float aV1)
	{
		float Delta = aV1 - aV0;
		if (Delta > mViewSize / 2)
			return Delta - mViewSize;
		if (Delta < -mViewSize / 2)
			return Delta + mViewSize;
		return Delta;
	}



	public void Render()
	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glScalef(2.0f / 1024, -2.0f / 768, 1);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		float CameraX = 0;
		float CameraY = 0;
		if (mCameraObject !is null)
		{
			CameraX = mCameraObject.X;
			CameraY = mCameraObject.Y;
		}

		// render game view
		mObelisk.SetLinearFiltering(true);
		glMatrixMode(GL_MODELVIEW);
		foreach (CViewObject Object; mViewObjects)
		{
			float X = Object.X - CameraX;
			float Y = Object.Y - CameraY;
			if (X < -mViewSize / 2)
				X += mViewSize;
			else if (X > mViewSize / 2)
				X -= mViewSize;
			if (Y < -mViewSize / 2)
				Y += mViewSize;
			if (Y > mViewSize / 2)
				Y -= mViewSize;
			// clip object if it is not in view
			//float MaxSize = sqrt(Object.XSize ^^ 2 + Object.YSize ^^ 2) / 2;
			float MaxSize = (Object.XSize + Object.YSize) / 2;
			if (X + MaxSize < -512 || X - MaxSize > 512 ||
					Y + MaxSize < -384 || Y - MaxSize > 384)
				continue;
			glPushMatrix();
			glTranslatef(X, Y, Object.Z);
			glRotatef(Object.Angle * 180 / PI, 0, 0, 1);
			auto v0 = new CVector(-Object.XSize / 2, -Object.YSize / 2, 0);
			auto v1 = new CVector(Object.XSize / 2, -Object.YSize / 2, 0);
			auto v2 = new CVector(Object.XSize / 2, Object.YSize / 2, 0);
			auto v3 = new CVector(-Object.XSize / 2, Object.YSize / 2, 0);
			final switch (Object.RenderMode)
			{
				case ERenderMode.Colour:
					mObelisk.TriColour(v0, v1, v2, Object.Colour);
					mObelisk.TriColour(v0, v2, v3, Object.Colour);
					break;
				case ERenderMode.Texture:
					auto Texture = mObelisk.GetTexture(Object.Texture);
					mObelisk.TriTexture(v0, 0, 0, v1, 1, 0, v2, 1, 1, Texture);
					mObelisk.TriTexture(v0, 0, 0, v2, 1, 1, v3, 0, 1, Texture);
					break;
				case ERenderMode.ShaderWithIntensity:
					mObelisk.ShaderUniform("Star.glsl", "uIntensity", Object.Intensity);
					mObelisk.QuadShader(v0, 0, 0, v1, 1, 0, v2, 1, 1, v3, 0, 1, Object.Shader);
					break;
			}
			if (mShowThrust && Object is mCameraObject)
			{
				auto Texture = mObelisk.GetTexture("Thrust.png");
				v0 = new CVector(-7 - 16, -4, 0);
				v1 = new CVector(0 - 16, -4, 0);
				v2 = new CVector(0 - 16, 4, 0);
				v3 = new CVector(-7 - 16, 4, 0);
				mObelisk.TriTexture(v0, 0, 0, v1, 1, 0, v2, 1, 1, Texture);
				mObelisk.TriTexture(v0, 0, 0, v2, 1, 1, v3, 0, 1, Texture);
			}
			glPopMatrix();
		}

		// render minimap view
		mObelisk.SetLinearFiltering(false);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glPushMatrix();
		glTranslatef(512 - 191, -384, 0);
		auto b0 = new CVector(0, 0, 0);
		auto b1 = new CVector(191, 0, 0);
		auto b2 = new CVector(191, 191, 0);
		auto b3 = new CVector(0, 191, 0);
		auto Texture = mObelisk.GetTexture("Minimap.png");
		mObelisk.TriTexture(b0, 0, 0, b1, 1, 0, b2, 1, 1, Texture);
		mObelisk.TriTexture(b0, 0, 0, b2, 1, 1, b3, 0, 1, Texture);
		glPopMatrix();

		const float cMinimapSize = 180;
		const float cMinimapSize2 = cMinimapSize / 2;
		glTranslatef(512 - cMinimapSize, -384, 0);
		if (mCameraObject !is null)
			glTranslatef(cMinimapSize2, cMinimapSize2, 0);
		foreach (CViewObject Object; mViewObjects)
		{
			if (Object.MinimapColour !is null)
			{
				float X = (Object.X - CameraX) * cMinimapSize / mViewSize;
				if (X < -cMinimapSize2)
					X += cMinimapSize;
				else if (X >= cMinimapSize2)
					X -= cMinimapSize;
				float Y = (Object.Y - CameraY) * cMinimapSize / mViewSize;
				if (Y < -cMinimapSize2)
					Y += cMinimapSize;
				else if (Y >= cMinimapSize2)
					Y -= cMinimapSize;
				glPushMatrix();
				glTranslatef(X, Y, 0);
				auto v0 = new CVector(0, 0, 0);
				auto v1 = new CVector(2, 0, 0);
				auto v2 = new CVector(2, 2, 0);
				auto v3 = new CVector(0, 2, 0);
				mObelisk.TriColour(v0, v1, v2, Object.MinimapColour);
				mObelisk.TriColour(v0, v2, v3, Object.MinimapColour);
				glPopMatrix();
			}
		}
	}
}