﻿module DUtil.CTextFileReader;
import std.algorithm, std.array, std.encoding, std.stream;

class CTextFileReader
{
	private enum EncodingType { Unknown, ANSI, UTF8, UTF16LE, UTF16BE }

	private Stream mStream;
	private EncodingType mEncoding = EncodingType.Unknown;
	private ubyte[] mData;
	private bool mEOF = false;
	private bool mFirstChar = true;
	private bool mHasNextChar = false;
	private dchar mNextChar;

	public this(Stream aStream)
	{
		mStream = aStream;
	}



	/**
	Reads the next character from the file.
	@return true if a character was read, or false if the end of the file was reached.
	*/
	public bool ReadChar(out dchar aChar)
	out (result)
	{
		assert(!result || (aChar >= 0 && aChar < 0x110000 && !(aChar >= 0xd800 && aChar < 0xe000)));
	}
	body
	{
		if (mEncoding == EncodingType.Unknown)
			DetectEncoding();
		dchar c;
		if (mHasNextChar)
		{
			mHasNextChar = false;
			c = mNextChar;
		}
		else
		{
			if (mData.length == 0 && mEOF)
				return false;
			if (!GetChar(c))
			{
				if (!mFirstChar)
				{
					aChar = '\n';   // always have a newline at the end of the last line
					return true;
				}
				return false;
			}
		}
		if (c == '\n')
		{
			aChar = c;
			mFirstChar = true;
		}
		else if (c == '\r')   // convert \r and \r\n newlines to \n
		{
			aChar = '\n';
			mFirstChar = true;
			if (GetChar(c))
			{
				if (c != '\n')
				{
					mHasNextChar = true;
					mNextChar = c;
				}
			}
		}
		else
		{
			aChar = c;
			mFirstChar = false;
		}
		return true;
	}



	private bool GetChar(out dchar aChar)
	out (result)
	{
		assert(!result || (aChar >= 0 && aChar < 0x110000 && !(aChar >= 0xd800 && aChar < 0xe000)));
	}
	body
	{
		final switch (mEncoding)
		{
			case EncodingType.Unknown:
				throw new Exception("Invalid encoding type");
			case EncodingType.ANSI:
				return ReadCharANSI(aChar);
			case EncodingType.UTF8:
				return ReadCharUTF8(aChar);
			case EncodingType.UTF16LE:
				return ReadCharUTF16LE(aChar);
			case EncodingType.UTF16BE:
				return ReadCharUTF16BE(aChar);
		}
	}



	private bool ReadCharANSI(out dchar aChar)
	{
		ubyte b;
		if (!GetByte(b))
			return false;
		Windows1252String ws = [ cast(Windows1252Char)b ];
		aChar = decode(ws);
		return true;
	}



	private bool ReadCharUTF8(out dchar aChar)
	{
		ubyte b;
		if (!GetByte(b))
			return false;
		if (b < 128)
		{
			aChar = b;
			return true;
		}
		if ((b & 0xe0) == 0xc0)
		{
			ubyte b1;
			if (!GetByte(b1))
				return false;
			aChar = ((b & 0x1f) << 6) | (b1 & 0x3f);
			assert(aChar >= 0x80 && aChar < 0x800);
			return true;
		}
		if ((b & 0xf0) == 0xe0)
		{
			ubyte b1;
			if (!GetByte(b1))
				return false;
			ubyte b2;
			if (!GetByte(b2))
				return false;
			aChar = ((b & 0xf) << 12) | ((b1 & 0x3f) << 6) | (b2 & 0x3f);
			assert(aChar >= 0x800 && aChar < 0x10000);
			if (aChar >= 0xd800 && aChar < 0xe000)
			{
				mEOF = true;
				return false;
			}
			return true;
		}
		if ((b & 0xf8) == 0xf0)
		{
			ubyte b1;
			if (!GetByte(b1))
				return false;
			ubyte b2;
			if (!GetByte(b2))
				return false;
			ubyte b3;
			if (!GetByte(b3))
				return false;
			aChar = ((b & 0x7) << 18) | ((b1 & 0x3f) << 12) | ((b2 & 0x3f) << 6) | (b3 & 0x3f);
			assert(aChar >= 0x10000 && aChar < 0x110000);
			return true;
		}
		mEOF = true;
		return false;
	}



	private bool ReadCharUTF16LE(out dchar aChar)
	{
		ubyte b0;
		if (!GetByte(b0))
			return false;
		ubyte b1;
		if (!GetByte(b1))
			return false;
		ushort c = (cast(ushort)b1 << 8) | b0;
		if (c >= 0xdc00 && c < 0xe000)
		{
			mEOF = true;
			return false;
		}
		if (!(c >= 0xd800 && c < 0xdc00))
		{
			aChar = c;
			return true;
		}
		if (!GetByte(b0))
			return false;
		if (!GetByte(b1))
			return false;
		ushort c1 = (cast(ushort)b1 << 8) | b0;
		if (!(c1 >= 0xdc00 && c1 < 0xe000))
		{
			mEOF = true;
			return false;
		}
		aChar = ((c & 0x3ff) << 10 | (c1 & 0x3ff)) + 0x10000;
		return true;
	}



	private bool ReadCharUTF16BE(out dchar aChar)
	{
		ubyte b1;
		if (!GetByte(b1))
			return false;
		ubyte b0;
		if (!GetByte(b0))
			return false;
		ushort c = (cast(ushort)b1 << 8) | b0;
		if (c >= 0xdc00 && c < 0xe000)
		{
			mEOF = true;
			return false;
		}
		if (!(c >= 0xd800 && c < 0xdc00))
		{
			aChar = c;
			return true;
		}
		if (!GetByte(b1))
			return false;
		if (!GetByte(b0))
			return false;
		ushort c1 = (cast(ushort)b1 << 8) | b0;
		if (!(c1 >= 0xdc00 && c1 < 0xe000))
		{
			mEOF = true;
			return false;
		}
		aChar = ((c & 0x3ff) << 10 | (c1 & 0x3ff)) + 0x10000;
		return true;
	}



	private bool GetByte(out ubyte aByte)
	{
		if (mData.length != 0)
		{
			aByte = mData[0];
			mData.popFront();
			return true;
		}
		if (mEOF)
			return false;
		return ReadFileByte(aByte);
	}



	private void DetectEncoding()
	out
	{
		assert(mEncoding != EncodingType.Unknown);
	}
	body
	{
		mEncoding = EncodingType.ANSI;
		ubyte b;
		if (!ReadFileByte(b))
			return;
		mData ~= [ b ];
		if (canFind([ 0xef, 0xfe, 0xff ], mData[0]))
		{
			if (!ReadFileByte(b))
				return;
			mData ~= [ b ];
			if ((mData[0] == 0xff && mData[1] == 0xfe) || (mData[0] == 0xfe && mData[1] == 0xff))
			{
				mEncoding = mData[0] == 0xff ? EncodingType.UTF16LE : EncodingType.UTF16BE;
				mData.length = 0;
				return;
			}
		}
		if (mData[0] == 0xef && mData[1] == 0xbb)
		{
			if (!ReadFileByte(b))
				return;
			mData ~= [ b ];
			if (mData[2] == 0xbf)
			{
				mEncoding = EncodingType.UTF8;
				mData.length = 0;
			}
		}
	}



	private bool ReadFileByte(out ubyte aByte)
	{
		try
		{
			mStream.read(aByte);
		}
		catch (ReadException)
		{
			mEOF = true;
			return false;
		}
		return true;
	}



	/*
	Test an empty stream contains no characters.
	*/
	unittest
	{
		ubyte[] Data = [];
		CTextFileReader Reader = new CTextFileReader(new MemoryStream(Data));
		dchar c;
		assert(!Reader.ReadChar(c));
	}



	/*
	Test that a \n is implicitly added if the last line does not end with \n.
	*/
	unittest
	{
		ubyte[] Data = [ 'A' ];
		CTextFileReader Reader = new CTextFileReader(new MemoryStream(Data));
		dchar c;
		assert(Reader.ReadChar(c) && c == 'A');
		assert(Reader.ReadChar(c) && c == '\n');
		assert(!Reader.ReadChar(c));
	}



	/*
	Test that no extra \n is added if the last line ends with \n.
	*/
	unittest
	{
		ubyte[] Data = [ 'A', '\n' ];
		CTextFileReader Reader = new CTextFileReader(new MemoryStream(Data));
		dchar c;
		assert(Reader.ReadChar(c) && c == 'A');
		assert(Reader.ReadChar(c) && c == '\n');
		assert(!Reader.ReadChar(c));
	}



	/*
	Test that \r\n is converted to \n.
	*/
	unittest
	{
		ubyte[] Data = [ '\r', '\n' ];
		CTextFileReader Reader = new CTextFileReader(new MemoryStream(Data));
		dchar c;
		assert(Reader.ReadChar(c) && c == '\n');
		assert(!Reader.ReadChar(c));
	}



	/*
	Test that \r followed by a character works.
	*/
	unittest
	{
		ubyte[] Data = [ '\r', 'A' ];
		CTextFileReader Reader = new CTextFileReader(new MemoryStream(Data));
		dchar c;
		assert(Reader.ReadChar(c) && c == '\n');
		assert(Reader.ReadChar(c) && c == 'A');
		assert(Reader.ReadChar(c) && c == '\n');
		assert(!Reader.ReadChar(c));
	}



	/*
	Test a multibyte code point in a UTF-8 encoded stream.
	*/
	unittest
	{
		ubyte[] Data = [ 0xef, 0xbb, 0xbf, 0xe2, 0x80, 0x94 ];
		CTextFileReader Reader = new CTextFileReader(new MemoryStream(Data));
		dchar c;
		assert(Reader.ReadChar(c) && c == 0x2014);
		assert(Reader.ReadChar(c) && c == '\n');
		assert(!Reader.ReadChar(c));
	}



	/*
	Test Windows-1252 character set.
	*/
	unittest
	{
		ubyte[] Data = [ 0x97 ];
		CTextFileReader Reader = new CTextFileReader(new MemoryStream(Data));
		dchar c;
		assert(Reader.ReadChar(c) && c == 0x2014);
		assert(Reader.ReadChar(c) && c == '\n');
		assert(!Reader.ReadChar(c));
	}
}