﻿module Windows.CApplication;
import std.ascii, std.conv, std.path, std.stream, std.string, std.utf, win32.windows,
		win32.winbase, win32.commctrl, win32.shlwapi;
import DOpenGL.gl, DOpenGL.glext, wgl = DOpenGL.wgl;
import PNGDecoder.CPNGDecoder, PNGDecoder.IImageDecoder;
import Obelisk.CColour, Obelisk.CTexture, Obelisk.CVector, Obelisk.EKey,
		Obelisk.IMainController, Obelisk.IObelisk, Obelisk.IThread, Shared.CObeliskException,
		Windows.CThreadWrapper;

extern (C) IMainController CreateMainController(IObelisk aObelisk);

extern (Windows)
{
	alias BOOL function(HANDLE hThread) FCancelSynchronousIo;
	alias HANDLE function(LPCWSTR, LPDWORD) FAvSetMmThreadCharacteristics;
	alias BOOL function(HANDLE) FAvRevertMmThreadCharacteristics;
}

PFNGLATTACHSHADERPROC glAttachShader;
PFNGLCOMPILESHADERPROC glCompileShader;
PFNGLCREATEPROGRAMPROC glCreateProgram;
PFNGLCREATESHADERPROC glCreateShader;
PFNGLDELETEPROGRAMPROC glDeleteProgram;
PFNGLDELETESHADERPROC glDeleteShader;
PFNGLGETPROGRAMINFOLOGPROC glGetProgramInfoLog;
PFNGLGETPROGRAMIVPROC glGetProgramiv;
PFNGLGETSHADERINFOLOGPROC glGetShaderInfoLog;
PFNGLGETSHADERIVPROC glGetShaderiv;
PFNGLGETUNIFORMLOCATIONPROC glGetUniformLocation;
PFNGLLINKPROGRAMPROC glLinkProgram;
PFNGLSHADERSOURCEPROC glShaderSource;
PFNGLUNIFORM1FPROC glUniform1f;
PFNGLUNIFORM2FPROC glUniform2f;
PFNGLUSEPROGRAMPROC glUseProgram;

void LoadOpenGLFunction(T)(string aName, ref T aFunction)
{
	PROC Proc = wglGetProcAddress(cast(char*)(aName ~ "\0"));
	if (Proc is NULL)
		throw new CObeliskException("Failed to load OpenGL function: " ~ aName);
	aFunction = cast(T)Proc;
}

class CApplication : IObelisk
{
	private HINSTANCE mInstance;
	private HMODULE mKernel32Dll = NULL;
	private HMODULE mAvrtDll = NULL;
	private FCancelSynchronousIo mCancelSynchronousIo;
	private FAvSetMmThreadCharacteristics mAvSetMmThreadCharacteristics;
	private FAvRevertMmThreadCharacteristics mAvRevertMmThreadCharacteristics;

	private IMainController mMainController;
	private bool mUseShaders = false;
	private bool mDirectorySet = false;
	private string mDirectory;
	private bool mCaptionSet = false;
	private string mCaption;
	private int mWindowXSize = -1;
	private int mWindowYSize;
	private HWND mWindow;
	private DWORD mMainThreadID;
	private CThreadWrapper[IThread] mThreads;
	private HANDLE mWaitableTimer = NULL;
	private bool mTimerStarted = false;
	private bool mTimerCallbackActive = false;
	private bool mMouseLeft;
	private bool mMouseLeftLast = false;
	private bool mMouseRight;
	private bool mMouseRightLast = false;
	private bool[256] mKeyPressed;
	private HDC mDC;
	private HGLRC mRenderingContext;
	private HGLRC mAltRenderingContext;
	private bool mAltAttached = false;
	debug bool mFrame = false;
	private bool mLinearFiltering = false;
	private bool mSpriteAlign = true;
	private float mAlpha = 1.0f;
	private CTexture[string] mTextures;
	private GLuint[string] mShaders;
	private GLuint[string] mPrograms;
	private CTexture[dchar][string] mFonts;
	private CTexture[dchar] mFont;
	private byte[][string] mSounds;
	private bool mRecord = false;

	public this(HINSTANCE aInstance)
	{
		mInstance = aInstance;
	}



	public bool CancelSynchronousIo(HANDLE aThread)
	{
		if (mCancelSynchronousIo == NULL)
			return false;
		return 0 != mCancelSynchronousIo(aThread);
	}



	public void Run()
	{
		mMainController = CreateMainController(this);
		Initialise();
		mMainController.Initialise();

		DWORD TaskIndex = 0;
		HANDLE Task = NULL;
		if (mAvrtDll != NULL)
		{
			Task = mAvSetMmThreadCharacteristics(cast(wchar*)"Games"w, &TaskIndex);
			assert(Task != cast(HANDLE)0);
		}

		HANDLE[1] WaitObjects;
		WaitObjects[0] = mWaitableTimer;
		while (true)
		{
			DWORD MWFMOResult = MsgWaitForMultipleObjects(mTimerStarted ? 1 : 0,
					cast(HANDLE*)WaitObjects, false, INFINITE, QS_ALLINPUT);
			assert(MWFMOResult == WAIT_OBJECT_0 || (mTimerStarted && MWFMOResult ==
					WAIT_OBJECT_0 + 1));
			if (mTimerStarted && MWFMOResult == WAIT_OBJECT_0)
				OnTimer();
			else
			{
				MSG Message;
				while (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
				{
					if (Message.message == WM_QUIT)
					{
						bool CHResult = 0 != CloseHandle(mWaitableTimer);
						assert(CHResult);
						if (mAvrtDll != NULL)
						{
							bool RMTCResult = 0 != mAvRevertMmThreadCharacteristics(Task);
							assert(RMTCResult);
						}
						delete mMainController;
						return;
					}
					DispatchMessage(&Message);
				}
			}
		}
	}



	private void Initialise()
	{
		if (!mCaptionSet)
			throw new CObeliskException("Program forgot to call SetCaption()");
		if (mWindowXSize == -1)
			throw new CObeliskException("Program forgot to call SetSize()");

		ProcessCommandLine();

		mMainThreadID = GetCurrentThreadId();

		SetDllDirectoryW(cast(wchar*)""w);
		mKernel32Dll = LoadLibraryW(cast(wchar*)"Kernel32.dll"w);
		if (mKernel32Dll == NULL)
			throw new Throwable("LoadLibrary(\"Kernel32.dll\") failed");
		mCancelSynchronousIo = cast(FCancelSynchronousIo)GetProcAddress(mKernel32Dll,
				"CancelSynchronousIo");
		mAvrtDll = LoadLibraryW(cast(wchar*)"Avrt.dll\0"w);
		if (mAvrtDll != NULL)
		{
			mAvSetMmThreadCharacteristics = cast(FAvSetMmThreadCharacteristics)GetProcAddress(
					mAvrtDll, "AvSetMmThreadCharacteristicsW\0");
			if (mAvSetMmThreadCharacteristics == NULL)
				throw new CObeliskException("Avrt.dll invalid: missing AvSetMmThreadCharacteristicsW()");
			mAvRevertMmThreadCharacteristics = cast(FAvRevertMmThreadCharacteristics)
					GetProcAddress(mAvrtDll, "AvRevertMmThreadCharacteristics");
			if (mAvRevertMmThreadCharacteristics == NULL)
				throw new CObeliskException("Avrt.dll invalid: missing AvRevertMmThreadCharacteristics()");
		}

		mWaitableTimer = CreateWaitableTimerW(NULL, false, NULL);
		assert(mWaitableTimer != NULL);

		INITCOMMONCONTROLSEX icc;
		bool ICCResult = 0 != InitCommonControlsEx(&icc);
		assert(ICCResult);

		WNDCLASSEXW ClassInfo;
		ClassInfo.cbSize = WNDCLASSEXW.sizeof;
		ClassInfo.style = CS_OWNDC;
		ClassInfo.lpfnWndProc = &MainWndProc;
		ClassInfo.hInstance = mInstance;
		ClassInfo.hCursor = LoadCursorW(NULL, IDC_ARROW);
		assert(ClassInfo.hCursor != NULL);
		ClassInfo.hbrBackground = cast(HBRUSH)GetStockObject(BLACK_BRUSH);
		ClassInfo.lpszClassName = cast(wchar*)"ObeliskMainWindow"w;
		ATOM RCResult = RegisterClassExW(&ClassInfo);
		assert(RCResult != 0);

		RECT r;
		bool SRResult = 0 != SetRect(&r, 0, 0, mWindowXSize, mWindowYSize);
		assert(SRResult);
		bool AWRResult = 0 != AdjustWindowRect(&r, WS_MINIMIZEBOX | WS_MAXIMIZEBOX |
				WS_POPUPWINDOW | WS_CAPTION, false);
		assert(AWRResult);
		int XSize = r.right - r.left;
		int X = (GetSystemMetrics(SM_CXSCREEN) - XSize) / 2;
		int YSize = r.bottom - r.top;
		int Y = (GetSystemMetrics(SM_CYSCREEN) - r.bottom) / 2;
		string Caption = mCaption;
		if (mRecord)
			Caption ~= " (RECORDING)";
		mWindow = CreateWindowExW(WS_EX_APPWINDOW, cast(wchar*)"ObeliskMainWindow"w,
				cast(wchar*)toUTF16z(Caption), WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_POPUPWINDOW |
				WS_CAPTION | WS_VISIBLE, X, Y, XSize, YSize, NULL, NULL, NULL, cast(LPVOID)this);
		assert(mWindow != NULL && GetLastError() == 0);



		PIXELFORMATDESCRIPTOR pfd;
		pfd.nSize = PIXELFORMATDESCRIPTOR.sizeof;
		pfd.nVersion = 1;
		pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
		pfd.iPixelType = PFD_TYPE_RGBA;
		pfd.cColorBits = 24;
		pfd.cDepthBits = 16;
		mDC = GetDC(mWindow);
		assert(mDC != NULL);
		int PixelFormat = ChoosePixelFormat(mDC, &pfd);
		if (PixelFormat == 0)
			throw new CObeliskException("ChoosePixelFormat() failed");
		// TODO: check that pixel format is good enough
		bool SPFResult = 0 != SetPixelFormat(mDC, PixelFormat, &pfd);
		if (!SPFResult)
			throw new Throwable("SetPixelFormat() failed");
		mRenderingContext = wgl.wglCreateContext(mDC);
		if (mRenderingContext == NULL)
			throw new Throwable("wglCreateContext() failed");
		mAltRenderingContext = wgl.wglCreateContext(mDC);
		if (mAltRenderingContext == NULL)
			throw new Throwable("wglCreateContext() failed (1)");
		bool SLResult = 0 != wgl.wglShareLists(mRenderingContext, mAltRenderingContext);
		assert(SLResult);
		bool WMCResult = 0 != wgl.wglMakeCurrent(mDC, mRenderingContext);
		assert(WMCResult && glGetError() == GL_NO_ERROR);

		if (mUseShaders)
		{
			GLubyte* VersionData = glGetString(GL_VERSION);
			assert(VersionData != NULL);
			for (int i = 0; ; i++)
			{
				if (VersionData[i] == 0 || !isDigit(VersionData[i]) || VersionData[i] == '.')
				{
					int MajorVersion = to!int(cast(string)VersionData[0 .. i]);
					if (MajorVersion < 2)
						throw new CObeliskException("OpenGL 2.0 or higher is required.");
					break;
				}
			}

			LoadOpenGLFunction("glAttachShader", glAttachShader);
			LoadOpenGLFunction("glCompileShader", glCompileShader);
			LoadOpenGLFunction("glCreateProgram", glCreateProgram);
			LoadOpenGLFunction("glCreateShader", glCreateShader);
			LoadOpenGLFunction("glDeleteProgram", glDeleteProgram);
			LoadOpenGLFunction("glDeleteShader", glDeleteShader);
			LoadOpenGLFunction("glGetProgramInfoLog", glGetProgramInfoLog);
			LoadOpenGLFunction("glGetProgramiv", glGetProgramiv);
			LoadOpenGLFunction("glGetShaderInfoLog", glGetShaderInfoLog);
			LoadOpenGLFunction("glGetShaderiv", glGetShaderiv);
			LoadOpenGLFunction("glGetUniformLocation", glGetUniformLocation);
			LoadOpenGLFunction("glLinkProgram", glLinkProgram);
			LoadOpenGLFunction("glShaderSource", glShaderSource);
			LoadOpenGLFunction("glUniform1f", glUniform1f);
			LoadOpenGLFunction("glUniform2f", glUniform2f);
			LoadOpenGLFunction("glUseProgram", glUseProgram);
		}



		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
		glClearDepth(1.0f);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LEQUAL);
		glEnable(GL_ALPHA_TEST);
		glAlphaFunc(GL_NOTEQUAL, 0.0f);
		glShadeModel(GL_SMOOTH);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
		SetSpriteViewport(mWindowXSize, mWindowYSize);
		assert(glGetError() == GL_NO_ERROR);
	}



	private void ProcessCommandLine()
	{
		//wchar *CommandLine = GetCommandLine();
		int NumArguments;
		LPWSTR *ArgumentsArray = CommandLineToArgvW(GetCommandLine(), &NumArguments);
		wstring[] Arguments;
		for (int i = 0; i < NumArguments; i++)
			Arguments ~= [ cast(wstring)ArgumentsArray[i][0 .. wcslen(ArgumentsArray[i])] ];
		LocalFree(cast(HANDLE)ArgumentsArray);
		foreach (wstring s; Arguments)
		{
			if (s == "-record"w)
				mRecord = true;
		}
	}



	private void WindowDestroyed()
	{
		if (mTimerStarted)
			StopTimer();
		foreach (CThreadWrapper Thread; mThreads)
		{
			Thread.Stop();
			delete Thread;
		}
		if (mUseShaders)
		{
			foreach (GLuint Program; mPrograms)
				glDeleteProgram(Program);
			foreach (GLuint Shader; mShaders)
				glDeleteShader(Shader);
		}
		bool Result = 0 != wgl.wglMakeCurrent(NULL, NULL);
		assert(Result);
		Result = 0 != wgl.wglDeleteContext(mRenderingContext);
		assert(Result);
		Result = 0 != wgl.wglDeleteContext(mAltRenderingContext);
		assert(Result);
		PlaySound(NULL, NULL, 0);
		PostQuitMessage(0);
	}



	private void OnTimer()
	{
		mMouseLeft = (GetAsyncKeyState(VK_LBUTTON) & (1 << 15)) != 0;
		mMouseRight = (GetAsyncKeyState(VK_RBUTTON) & (1 << 15)) != 0;
		mTimerCallbackActive = true;
		mMainController.OnTimer();
		mTimerCallbackActive = false;
		mMouseLeftLast = mMouseLeft;
		mMouseRightLast = mMouseRight;
		mKeyPressed[] = false;
	}



	private void OnKeyDown(uint aVirtualKey)
	{
		EKey Key;
		if (GetKey(aVirtualKey, Key))
			mMainController.OnKeyDown(Key);
		if (aVirtualKey <= 255)
		mKeyPressed[aVirtualKey] = true;
	}



	private void OnKeyUp(uint aVirtualKey)
	{
		EKey Key;
		if (GetKey(aVirtualKey, Key))
			mMainController.OnKeyUp(Key);
	}



	private bool GetKey(uint aVirtualKey, out EKey aKey)
	{
		switch (aVirtualKey)
		{
			case VK_LEFT:
				aKey = EKey.Left;
				return true;
			case VK_RIGHT:
				aKey = EKey.Right;
				return true;
			case VK_UP:
				aKey = EKey.Up;
				return true;
			case VK_DOWN:
				aKey = EKey.Down;
				return true;
			case VK_RETURN:
				aKey = EKey.Enter;
				return true;
			case VK_SPACE:
				aKey = EKey.Space;
				return true;
			case VK_ESCAPE:
				aKey = EKey.Escape;
				return true;
			case VK_BACK:
				aKey = EKey.Backspace;
				return true;
			case VK_TAB:
				aKey = EKey.Tab;
				return true;
			case VK_SHIFT:
			case VK_LSHIFT:
				aKey = EKey.LeftShift;
				return true;
			case VK_RSHIFT:
				aKey = EKey.RightShift;
				return true;
			case VK_CONTROL:
			case VK_LCONTROL:
				aKey = EKey.LeftControl;
				return true;
			case VK_RCONTROL:
				aKey = EKey.RightControl;
				return true;
			case VK_MENU:
			case VK_LMENU:
				aKey = EKey.LeftAlt;
				return true;
			case VK_RMENU:
				aKey = EKey.RightAlt;
				return true;
			case cast(uint)'A':
				aKey = EKey.A;
				return true;
			case cast(uint)'B':
				aKey = EKey.B;
				return true;
			case cast(uint)'C':
				aKey = EKey.C;
				return true;
			case cast(uint)'D':
				aKey = EKey.D;
				return true;
			case cast(uint)'E':
				aKey = EKey.E;
				return true;
			case cast(uint)'F':
				aKey = EKey.F;
				return true;
			case cast(uint)'G':
				aKey = EKey.G;
				return true;
			case cast(uint)'H':
				aKey = EKey.H;
				return true;
			case cast(uint)'I':
				aKey = EKey.I;
				return true;
			case cast(uint)'J':
				aKey = EKey.J;
				return true;
			case cast(uint)'K':
				aKey = EKey.K;
				return true;
			case cast(uint)'L':
				aKey = EKey.L;
				return true;
			case cast(uint)'M':
				aKey = EKey.M;
				return true;
			case cast(uint)'N':
				aKey = EKey.N;
				return true;
			case cast(uint)'O':
				aKey = EKey.O;
				return true;
			case cast(uint)'P':
				aKey = EKey.P;
				return true;
			case cast(uint)'Q':
				aKey = EKey.Q;
				return true;
			case cast(uint)'R':
				aKey = EKey.R;
				return true;
			case cast(uint)'S':
				aKey = EKey.S;
				return true;
			case cast(uint)'T':
				aKey = EKey.T;
				return true;
			case cast(uint)'U':
				aKey = EKey.U;
				return true;
			case cast(uint)'V':
				aKey = EKey.V;
				return true;
			case cast(uint)'W':
				aKey = EKey.W;
				return true;
			case cast(uint)'X':
				aKey = EKey.X;
				return true;
			case cast(uint)'Y':
				aKey = EKey.Y;
				return true;
			case cast(uint)'Z':
				aKey = EKey.Z;
				return true;
			case cast(uint)'0':
				aKey = EKey.k0;
				return true;
			case cast(uint)'1':
				aKey = EKey.k1;
				return true;
			case cast(uint)'2':
				aKey = EKey.k2;
				return true;
			case cast(uint)'3':
				aKey = EKey.k3;
				return true;
			case cast(uint)'4':
				aKey = EKey.k4;
				return true;
			case cast(uint)'5':
				aKey = EKey.k5;
				return true;
			case cast(uint)'6':
				aKey = EKey.k6;
				return true;
			case cast(uint)'7':
				aKey = EKey.k7;
				return true;
			case cast(uint)'8':
				aKey = EKey.k8;
				return true;
			case cast(uint)'9':
				aKey = EKey.k9;
				return true;
			default:
				return false;
		}
	}



	private uint GetVirtualKey(EKey aKey)
	{
		switch (aKey)
		{
			case EKey.Left:
				return VK_LEFT;
			case EKey.Right:
				return VK_RIGHT;
			case EKey.Up:
				return VK_UP;
			case EKey.Down:
				return VK_DOWN;
			case EKey.Enter:
				return VK_RETURN;
			case EKey.Space:
				return VK_SPACE;
			case EKey.Escape:
				return VK_ESCAPE;
			case EKey.Backspace:
				return VK_BACK;
			case EKey.Tab:
				return VK_TAB;
			case EKey.LeftShift:
				return VK_LSHIFT;
			case EKey.RightShift:
				return VK_RSHIFT;
			case EKey.LeftControl:
				return VK_LCONTROL;
			case EKey.RightControl:
				return VK_RCONTROL;
			case EKey.LeftAlt:
				return VK_LMENU;
			case EKey.RightAlt:
				return VK_RMENU;
			case EKey.A:
				return cast(uint)'A';
			case EKey.B:
				return cast(uint)'B';
			case EKey.C:
				return cast(uint)'C';
			case EKey.D:
				return cast(uint)'D';
			case EKey.E:
				return cast(uint)'E';
			case EKey.F:
				return cast(uint)'F';
			case EKey.G:
				return cast(uint)'G';
			case EKey.H:
				return cast(uint)'H';
			case EKey.I:
				return cast(uint)'I';
			case EKey.J:
				return cast(uint)'J';
			case EKey.K:
				return cast(uint)'K';
			case EKey.L:
				return cast(uint)'L';
			case EKey.M:
				return cast(uint)'M';
			case EKey.N:
				return cast(uint)'N';
			case EKey.O:
				return cast(uint)'O';
			case EKey.P:
				return cast(uint)'P';
			case EKey.Q:
				return cast(uint)'Q';
			case EKey.R:
				return cast(uint)'R';
			case EKey.S:
				return cast(uint)'S';
			case EKey.T:
				return cast(uint)'T';
			case EKey.U:
				return cast(uint)'U';
			case EKey.V:
				return cast(uint)'V';
			case EKey.W:
				return cast(uint)'W';
			case EKey.X:
				return cast(uint)'X';
			case EKey.Y:
				return cast(uint)'Y';
			case EKey.Z:
				return cast(uint)'Z';
			case EKey.k0:
				return cast(uint)'0';
			case EKey.k1:
				return cast(uint)'1';
			case EKey.k2:
				return cast(uint)'2';
			case EKey.k3:
				return cast(uint)'3';
			case EKey.k4:
				return cast(uint)'4';
			case EKey.k5:
				return cast(uint)'5';
			case EKey.k6:
				return cast(uint)'6';
			case EKey.k7:
				return cast(uint)'7';
			case EKey.k8:
				return cast(uint)'8';
			case EKey.k9:
				return cast(uint)'9';
			default:
				throw new Exception("Invalid key");
		}
	}



	public bool OnSetCursor()
	{
		int x, y;
		GetCursorPos(x, y);
		RECT Client;
		bool GCRResult = 0 != GetClientRect(mWindow, &Client);
		assert(GCRResult);
		if (x >= 0 && x < Client.right - Client.left && y >= 0 && y < Client.bottom - Client.top)
		{
			SetCursor(NULL);
			return true;
		}
		return false;
	}



	private wstring GetDataFilePath(string aFileName)
	{
		DWORD Size = MAX_PATH;
		wchar[] FilePath = new wchar [Size + 1];
		while (true)
		{
			DWORD Result = GetModuleFileName(NULL, cast(wchar*)FilePath, Size + 1);
			assert(Result != 0 && Result <= Size);
			if (Result == Size)
			{
				Size *= 2;
				FilePath = new wchar [Size + 1];
			}
			else
				break;
		}
		bool PRFSResult = 0 != PathRemoveFileSpecW(cast(wchar*)FilePath);
		assert(PRFSResult);
		sizediff_t Index = indexOf(FilePath, '\0');
		assert(Index != -1);
		return cast(wstring)(FilePath[0 .. Index]) ~ "\\Data\\"w ~ toUTF16(aFileName) ~ "\0"w;
	}



	public void SetDirectory(string aDirectory)
	{
		assert(!mDirectorySet);   // TODO: disallow \ and /
		mDirectorySet = true;
		mDirectory = aDirectory;
	}



	public void SetCaption(string aCaption)
	{
		assert(!mCaptionSet);
		mCaptionSet = true;
		mCaption = aCaption;
	}



	public void SetSize(int aXSize, int aYSize)
	{
		assert(mWindowXSize == -1 && aXSize > 0 && aYSize > 0);
		mWindowXSize = aXSize;
		mWindowYSize = aYSize;
	}



	public void SetShaders(bool aUseShaders)
	{
		mUseShaders = aUseShaders;
	}



	public void StartTimer()
	{
		assert(!mTimerStarted);
		LARGE_INTEGER DueTime;
		DueTime.QuadPart = -1;
		bool SWTResult = 0 != SetWaitableTimer(mWaitableTimer, &DueTime, 20, NULL, NULL, false);
		assert(SWTResult);
		mTimerStarted = true;
	}



	public void StopTimer()
	{
		assert(mTimerStarted);
		bool Result = 0 != CancelWaitableTimer(mWaitableTimer);
		assert(Result);
		mTimerStarted = false;
	}



	public uint CreateTimer()
	{
		return GetTickCount();
	}



	public uint Duration(uint aTimer)
	{
		return GetTickCount() - aTimer;
	}



	public bool MouseButtonDown(bool aRight)
	{
		assert(mTimerCallbackActive);
		return aRight ? mMouseRight : mMouseLeft;
	}



	public bool MouseButtonClicked(bool aRight)
	{
		assert(mTimerCallbackActive);
		return aRight ? mMouseRight && !mMouseRightLast : mMouseLeft && !mMouseLeftLast;
	}



	public bool MouseButtonReleased(bool aRight)
	{
		assert(mTimerCallbackActive);
		return aRight ? !mMouseRight && mMouseRightLast : !mMouseLeft && mMouseLeftLast;
	}



	public void GetCursorPos(out int aCursorX, out int aCursorY)
	{
		POINT Cursor;
		bool GCPResult = 0 != win32.winuser.GetCursorPos(&Cursor);
		assert(GCPResult);
		bool STCResult = 0 != ScreenToClient(mWindow, &Cursor);
		assert(STCResult);
		aCursorX = Cursor.x;
		aCursorY = Cursor.y;
	}



	public bool IsKeyPressed(EKey aKey)
	{
		assert(mTimerCallbackActive);
		uint VirtualKey = GetVirtualKey(aKey);
		return VirtualKey > 255 ? false : mKeyPressed[VirtualKey];
	}



	public bool IsKeyDown(EKey aKey)
	{
		if (GetFocus() == NULL)
			return false;
		uint VirtualKey = GetVirtualKey(aKey);
		if ((GetAsyncKeyState(VirtualKey) & ~1) != 0)
			return true;
		return VirtualKey > 255 ? false : mKeyPressed[VirtualKey];
	}



	private int NextPowerOf2(int n)
	{
		int Next = 1;
		while (n > Next)
			Next <<= 1;
		return Next;
	}



	public void LoadTexture(string aName)
	{
		wstring TexturePath = GetDataFilePath(aName);
		File f;
		try
		{
			f = new File(to!string(TexturePath), FileMode.In);
		}
		catch (OpenException oe)
		{
			throw new CObeliskException("Image file not found: " ~ aName);
		}
		try
		{
			IImageDecoder Decoder;
			string Extension = toUpper(getExt(aName));
			if (cmp(Extension, "PNG") == 0)
				Decoder = new CPNGDecoder(f);
			else
				throw new CObeliskException("Unknown image format: " ~ aName);
			if (!Decoder.IsValid())
				throw new CObeliskException("Image file invalid: " ~ aName);
			int XSize2 = NextPowerOf2(Decoder.GetXSize());
			int YSize2 = NextPowerOf2(Decoder.GetYSize());
			uint[] Data = new uint [XSize2 * YSize2];
			for (int y = 0; y < Decoder.GetYSize(); y++)
			{
				for (int x = 0; x < Decoder.GetXSize(); x++)
				{
					if (!Decoder.GetPixel(Data[x + y * XSize2]))
						throw new CObeliskException("Image file invalid: " ~ aName);
				}
			}
			GLuint Name;
			glGenTextures(1, &Name);
			glBindTexture(GL_TEXTURE_2D, Name);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, XSize2, YSize2, 0, GL_RGBA, GL_UNSIGNED_BYTE,
					Data.ptr);
			assert(glGetError() == GL_NO_ERROR);
			mTextures[aName] = new CTexture(Decoder.GetXSize(), Decoder.GetYSize(),
					cast(float)Decoder.GetXSize() / XSize2, cast(float)Decoder.GetYSize() /
					YSize2, Name);
		}
		finally
		{
			f.close();
		}
	}



	public CTexture CreateTexture(string aName, uint[] aData, int aXSize, int aYSize)
	{
		int XSize2 = NextPowerOf2(aXSize);
		int YSize2 = NextPowerOf2(aYSize);
		uint[] Data = new uint [XSize2 * YSize2];
		int Index = 0;
		for (int y = 0; y < aYSize; y++)
		{
			for (int x = 0; x < aXSize; x++)
				Data[y * XSize2 + x] = aData[Index++];
		}

		GLuint Name;
		glGenTextures(1, &Name);
		glBindTexture(GL_TEXTURE_2D, Name);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, XSize2, YSize2, 0, GL_RGBA, GL_UNSIGNED_BYTE,
				cast(void*)Data);
		assert(glGetError() == GL_NO_ERROR);
		CTexture Texture = new CTexture(aXSize, aYSize, cast(float)aXSize / XSize2,
				cast(float)aYSize / YSize2, Name);
		mTextures[aName] = Texture;
		return Texture;
	}



	public CTexture GetTexture(string aName)
	{
		return mTextures[aName];
	}



	public void LoadShader(string aName)
	{
		if (!mUseShaders)
			throw new CObeliskException("Shaders need to be enabled with SetShaders()");
		wstring FilePath = GetDataFilePath(aName);
		HANDLE File = CreateFile(cast(wchar*)FilePath, GENERIC_READ, FILE_SHARE_READ, NULL,
				OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		if (File == INVALID_HANDLE_VALUE)
			throw new CObeliskException("Shader file not found: " ~ aName);
		LARGE_INTEGER Size;
		bool FSResult = 0 != GetFileSizeEx(File, &Size);
		assert(FSResult);
		if (Size.HighPart != 0)
			throw new CObeliskException("Shader file too large (> 4GiB): " ~ aName);
		byte[] Data = new byte [Size.LowPart + 1];
		DWORD BytesRead;
		if (!ReadFile(File, Data.ptr, Size.LowPart, &BytesRead, NULL))
			assert(GetLastError() == ERROR_OPERATION_ABORTED);
		else
			assert(BytesRead == Size.LowPart);
		bool CHResult = 0 != CloseHandle(File);
		assert(CHResult);

		GLuint Shader = glCreateShader(GL_FRAGMENT_SHADER);
		glShaderSource(Shader, 1, cast(GLchar**)[ Data.ptr ].ptr, NULL);
		glCompileShader(Shader);
		GLint CompileStatus;
		glGetShaderiv(Shader, GL_COMPILE_STATUS, &CompileStatus);
		if (CompileStatus == GL_FALSE)
			throw new CObeliskException("Shader compile failed: " ~ aName);
		mShaders[aName] = Shader;
		GLuint Program = glCreateProgram();
		glAttachShader(Program, Shader);
		glLinkProgram(Program);
		GLint LinkStatus;
		glGetProgramiv(Program, GL_LINK_STATUS, &LinkStatus);
		if (LinkStatus == GL_FALSE)
			throw new CObeliskException("Program link failed for shader: " ~ aName);
		mPrograms[aName] = Program;
		assert(glGetError() == GL_NO_ERROR);
	}



	public void SetLinearFiltering(bool aLinearFiltering)
	{
		mLinearFiltering = aLinearFiltering;
	}



	public void SetSpriteAlign(bool aSpriteAlign)
	{
		mSpriteAlign = aSpriteAlign;
	}



	public void BeginFrame()
	{
		debug
		{
			assert(!mFrame);
			mFrame = true;
		}
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		assert(glGetError() == GL_NO_ERROR);
	}



	public void EndFrame()
	{
		debug
		{
			assert(mFrame);
			mFrame = false;
		}
		bool SBResult = 0 != SwapBuffers(mDC);
		assert(SBResult);
	}



	public void SetBackgroundColour(float aR, float aG, float aB)
	{
		assert(aR >= 0 && aR <= 1 && aG >= 0 && aG <= 1 && aB >= 0 && aB <= 1);
		glClearColor(aR, aG, aB, 1.0f);
	}



	public void SetAlpha(float aAlpha)
	{
		mAlpha = aAlpha;
	}



	public void SetSpriteViewport(int aXSize, int aYSize)
	{
		GLenum MatrixMode;
		glGetIntegerv(GL_MATRIX_MODE, cast(GLint*)&MatrixMode);
		glMatrixMode(GL_PROJECTION);
		glTranslatef(-1, 1, 0);
		glScalef(2.0f / aXSize, -2.0f / aYSize, 1);
		glMatrixMode(MatrixMode);
	}



	public void DrawSprite(CTexture aTexture, int aX, int aY, int aXSize = -1, int aYSize = -1,
			float aXAlign = 0, float aYAlign = 0)
	{
		debug assert(mFrame);

		int XSize = aXSize == -1 ? aTexture.GetXSize() : aXSize;
		int YSize = aYSize == -1 ? aTexture.GetYSize() : aYSize;
		float X = mSpriteAlign ? cast(float)(aX - cast(int)(aXAlign * XSize)) :
				aX - aXAlign * XSize;
		float Y = mSpriteAlign ? cast(float)(aY - cast(int)(aYAlign * YSize)) :
				aY - aYAlign * YSize;

		glEnable(GL_TEXTURE_2D);
		glDisable(GL_CULL_FACE);
		glBindTexture(GL_TEXTURE_2D, aTexture.GetTextureName());
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
		glBegin(GL_QUADS);
		glTexCoord2f(0, 0);
		glColor4f(1, 1, 1, mAlpha);
		glVertex2f(X, Y);
		glTexCoord2f(aTexture.GetXRatio(), 0);
		glColor4f(1, 1, 1, mAlpha);
		glVertex2f(X + XSize, Y);
		glTexCoord2f(aTexture.GetXRatio(), aTexture.GetYRatio());
		glColor4f(1, 1, 1, mAlpha);
		glVertex2f(X + XSize, Y + YSize);
		glTexCoord2f(0, aTexture.GetYRatio());
		glColor4f(1, 1, 1, mAlpha);
		glVertex2f(X, Y + YSize);
		glEnd();
		assert(glGetError() == GL_NO_ERROR);
	}



	public void TriColour(CVector aV0, CVector aV1, CVector aV2, CColour aColour)
	{
		TriColour(aV0, aColour, aV1, aColour, aV2, aColour);
	}

	public void TriColour(CVector aV0, CColour aC0, CVector aV1, CColour aC1, CVector aV2,
			CColour aC2)
	{
		debug assert(mFrame);

		glDisable(GL_TEXTURE_2D);
		glEnable(GL_CULL_FACE);
		glBegin(GL_TRIANGLES);
		glColor4f(aC0.R, aC0.G, aC0.B, mAlpha);
		glVertex3f(aV0.X, aV0.Y, aV0.Z);
		glColor4f(aC1.R, aC1.G, aC1.B, mAlpha);
		glVertex3f(aV1.X, aV1.Y, aV1.Z);
		glColor4f(aC2.R, aC2.G, aC2.B, mAlpha);
		glVertex3f(aV2.X, aV2.Y, aV2.Z);
		glEnd();
		assert(glGetError() == GL_NO_ERROR);
	}



	public void TriTexture(CVector aV0, float aS0, float aT0, CVector aV1, float aS1,
			float aT1, CVector aV2, float aS2, float aT2, CTexture aTexture)
	{
		debug assert(mFrame);

		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, aTexture.GetTextureName());
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
		glEnable(GL_CULL_FACE);
		glBegin(GL_TRIANGLES);
		glTexCoord2f(aS0 * aTexture.GetXRatio(), aT0 * aTexture.GetYRatio());
		glColor4f(1, 1, 1, mAlpha);
		glVertex3f(aV0.X, aV0.Y, aV0.Z);
		glTexCoord2f(aS1 * aTexture.GetXRatio(), aT1 * aTexture.GetYRatio());
		glColor4f(1, 1, 1, mAlpha);
		glVertex3f(aV1.X, aV1.Y, aV1.Z);
		glTexCoord2f(aS2 * aTexture.GetXRatio(), aT2 * aTexture.GetYRatio());
		glColor4f(1, 1, 1, mAlpha);
		glVertex3f(aV2.X, aV2.Y, aV2.Z);
		glEnd();
		assert(glGetError() == GL_NO_ERROR);
	}



	public void QuadShader(CVector aV0, float aS0, float aT0, CVector aV1, float aS1,
			float aT1, CVector aV2, float aS2, float aT2, CVector aV3, float aS3, float aT3,
			string aShaderName)
	{
		glUseProgram(mPrograms[aShaderName]);
		glEnable(GL_CULL_FACE);
		glBegin(GL_QUADS);
		glTexCoord2f(aS0, aT0);
		glVertex3f(aV0.X, aV0.Y, aV0.Z);
		glTexCoord2f(aS1, aT1);
		glVertex3f(aV1.X, aV1.Y, aV1.Z);
		glTexCoord2f(aS2, aT2);
		glVertex3f(aV2.X, aV2.Y, aV2.Z);
		glTexCoord2f(aS3, aT3);
		glVertex3f(aV3.X, aV3.Y, aV3.Z);
		glEnd();
		glUseProgram(0);
		assert(glGetError() == GL_NO_ERROR);
	}



	public void ShaderUniform(string aShaderName, string aUniform, float aValue)
	{
		string UniformStr = aUniform ~ "\0";
		GLint Location = glGetUniformLocation(mPrograms[aShaderName], cast(GLchar*)(UniformStr.ptr));
		glUseProgram(mPrograms[aShaderName]);
		glUniform1f(Location, aValue);
		glUseProgram(0);
	}



	public void LoadFontImage(string aName)
	{
		wstring FontPath = GetDataFilePath(aName);
		File f;
		try
		{
			f = new File(to!string(FontPath), FileMode.In);
		}
		catch (OpenException oe)
		{
			throw new CObeliskException("Font file not found: " ~ aName);
		}
		uint[] Data;
		int XSize;
		int YSize;
		try
		{
			IImageDecoder Decoder;
			string Extension = toUpper(getExt(aName));
			if (cmp(Extension, "PNG") == 0)
				Decoder = new CPNGDecoder(f, 0xff, 10000);
			else
				throw new CObeliskException("Unknown image format: " ~ aName);
			if (!Decoder.IsValid())
				throw new CObeliskException("Font file invalid: " ~ aName);
			XSize = Decoder.GetXSize();
			YSize = Decoder.GetYSize();
			Data = new uint [XSize * YSize];
			for (int y = 0; y < YSize; y++)
			{
				for (int x = 0; x < XSize; x++)
				{
					if (!Decoder.GetPixel(Data[x + y * XSize]))
						throw new CObeliskException("Font file invalid: " ~ aName);
				}
			}
		}
		finally
		{
			f.close();
		}

		int GlyphYSize2 = NextPowerOf2(YSize);
		int XPos = 0;
		CTexture[dchar] Font;
		for (int i = 32; i <= 126; i++)
		{
			int GlyphXSize = 0;
			for (int x = XPos; x < XSize; x++)
			{
				if (Data[x] == 0xff00ff00)   // green: character edge
				{
					GlyphXSize = x - XPos;
					break;
				}
			}
			int GlyphXSize2 = NextPowerOf2(GlyphXSize);
			uint[] GlyphData = new uint [GlyphXSize2 * GlyphYSize2];
			for (int y = 0; y < YSize; y++)
			{
				for (int x = 0; x < GlyphXSize; x++)
				{
					uint Colour = Data[(x + XPos) + y * XSize];
					if (Colour == 0xffff00ff)   // magenta: transparent
						GlyphData[x + y * GlyphXSize2] = 0x00000000;
					else
						GlyphData[x + y * GlyphXSize2] = 0xff000000 | Colour;
				}
			}
			XPos += GlyphXSize + 1;

			GLuint Name;
			glGenTextures(1, &Name);
			glBindTexture(GL_TEXTURE_2D, Name);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, GlyphXSize2, GlyphYSize2, 0, GL_RGBA,
					GL_UNSIGNED_BYTE, GlyphData.ptr);
			assert(glGetError() == GL_NO_ERROR);
			Font[cast(dchar)i] = new CTexture(GlyphXSize, YSize, cast(float)GlyphXSize /
					GlyphXSize2, cast(float)YSize / GlyphYSize2, Name);
		}
		mFonts[aName] = Font;
	}



	public void SetFont(string aName)
	{
		mFont = mFonts[aName];
	}



	public void DrawText(string aText, int aX, int aY)
	{
		int X = aX;
		size_t i = 0;
		while (i < aText.length)
		{
			CTexture Texture = mFont[decode(aText, i)];
			DrawSprite(Texture, X, aY);
			X += Texture.GetXSize();
		}
	}



	public void LoadSound(string aName)
	{
		wstring FilePath = GetDataFilePath(aName);
		HANDLE File = CreateFile(cast(wchar*)FilePath, GENERIC_READ, FILE_SHARE_READ, NULL,
				OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
		if (File == INVALID_HANDLE_VALUE)
			throw new CObeliskException("Sound file not found: " ~ aName);
		LARGE_INTEGER Size;
		bool FSResult = 0 != GetFileSizeEx(File, &Size);
		assert(FSResult);
		if (Size.HighPart != 0)
			throw new CObeliskException("Sound file too large (> 4GiB): " ~ aName);
		byte[] Data = new byte [Size.LowPart];
		DWORD BytesRead;
		if (!ReadFile(File, Data.ptr, Size.LowPart, &BytesRead, NULL))
			assert(GetLastError() == ERROR_OPERATION_ABORTED);
		else
		{
			assert(BytesRead == Size.LowPart);
			mSounds[aName] = Data;
		}
		bool CHResult = 0 != CloseHandle(File);
		assert(CHResult);
	}



	public void Sound(string aName)
	{
		PlaySound(cast(LPCTSTR)mSounds[aName], NULL, SND_MEMORY | SND_ASYNC | SND_NODEFAULT);
	}
}



extern (Windows)
LRESULT MainWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
		case WM_CREATE:
			CREATESTRUCT *cs = cast(CREATESTRUCT*)lParam;
			SetLastError(0);
			LONG_PTR Result = SetWindowLongPtrW(hwnd, GWLP_USERDATA, cast(LONG_PTR)cs.lpCreateParams);
			if (Result == 0 && GetLastError() != 0)
				throw new Exception("SetWindowLongPtr() failed");
			return 0;
		case WM_DESTROY:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CApplication Application = cast(CApplication)(cast(LPVOID)Pointer);
			Application.WindowDestroyed();
			return 0;
		case WM_CLOSE:
			DestroyWindow(hwnd);
			return 0;
		case WM_KEYDOWN:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CApplication Application = cast(CApplication)(cast(LPVOID)Pointer);
			Application.OnKeyDown(cast(uint)wParam);
			return 0;
		case WM_KEYUP:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CApplication Application = cast(CApplication)(cast(LPVOID)Pointer);
			Application.OnKeyUp(cast(uint)wParam);
			return 0;
		case WM_SETCURSOR:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CApplication Application = cast(CApplication)(cast(LPVOID)Pointer);
			if (Application.OnSetCursor())
				return TRUE;
			break;
		case WM_SETFOCUS:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CApplication Application = cast(CApplication)(cast(LPVOID)Pointer);
			if (Application != NULL)
				Application.mMainController.OnActivate();
			break;
		case WM_KILLFOCUS:
			LONG_PTR Pointer = GetWindowLongPtr(hwnd, GWLP_USERDATA);
			CApplication Application = cast(CApplication)(cast(LPVOID)Pointer);
			Application.mMainController.OnDeactivate();
			break;
		default:
			break;
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}